{****************************************************************************}
{*                                                                          *}
{*                              Program/Module                              *}
{*                                   from                                   *}
{*                 NUMERICAL TOOLBOX FOR VERIFIED COMPUTING I               *}
{*                         Basic Numerical Problems                         *}
{*                                                                          *}
{*      Copyright (c) 1993   Rolf Hammer, Matthias Hocks, Dietmar Ratz      *}
{*                                                                          *}
{* This program/module is free software for non-commercial use. For details *}
{* on theory, algorithms, and programs, see the book                        *}
{*                                                                          *}
{*  R. Hammer, M. Hocks, U. Kulisch, D. Ratz:  Numerical Toolbox for        *}
{*  Verified Computing I - Basic Numerical Problems. Springer-Verlag,       *}
{*  Heidelberg (ISBN 3-540-57118-3), New York (ISBN 0-387-57118-3), 1993.   *}
{*                                                                          *}
{* This program/module is distributed WITHOUT ANY WARRANTY. For details,    *}
{* see the "Disclaimer / Legal Matters" of the book (page iv).              *}
{*                                                                          *}
{****************************************************************************}
{----------------------------------------------------------------------------}
{ Purpose: Computation of a verified solution of a square linear system of   }
{    equations A*x = b with full real matrix A and real right-hand side b.   }
{ Method: Transformation of A*x = b to fixed-point form and applying an      }
{    interval residual iteration.                                            }
{ Global functions and procedures:                                           }
{   procedure LinSolve(...)      : To get a verified enclosure of the        }
{                                  solution, two versions                    }
{   function LinSolveErrMsg(...) : To get an error message text              }
{----------------------------------------------------------------------------}
module linsys;      { Linear system solving }
                    { ---    ---            }
use
  mv_ari,           { Matrix/vector arithmetic          }
  mvi_ari,          { Matrix/vector interval arithmetic }
  matinv;           { Matrix inversion                  }

{----------------------------------------------------------------------------}
{ Error codes used in this module.                                           }
{----------------------------------------------------------------------------}
const
  NoError      = 0;   { No error occurred.                        }
  NotSquare    = 1;   { System to be solved is not square.        }
  DimensionErr = 2;   { Dimensions of A*x = b are not compatible. }
  InvFailed    = 3;   { System is probably singular.              }
  VerifFailed  = 4;   { Verification failed, system is probably   }
                      { ill-conditioned.                          }

{----------------------------------------------------------------------------}
{ Error messages depending on the error code.                                }
{----------------------------------------------------------------------------}
global function LinSolveErrMsg ( Err : integer ) : string;
var
  Msg : string;
begin
  case Err of
    NoError     : Msg := '';
    NotSquare   : Msg := 'System to be solved is not square';
    DimensionErr: Msg := 'Dimensions of A*x = b are not compatible';
    InvFailed   : Msg := 'System is probably singular';
    VerifFailed : Msg := 'Verification failed, system is probably ' +
                         'ill-conditioned';
    else        : Msg := 'Code not defined';
  end;
  if (Err <> NoError) then Msg := 'Error: ' + Msg + '!';
  LinSolveErrMsg := Msg;
end;

{----------------------------------------------------------------------------}
{ Computes the absolute value of a real vector component by component.       }
{----------------------------------------------------------------------------}
function abs( v : rvector ) : rvector[lb(v)..ub(v)];
var
  i : integer;
begin
  for i := lb(v) to ub(v) do abs[i] := abs(v[i]);
end;


{----------------------------------------------------------------------------}
{ Computes an upper bound for the maximum norm of a real matrix 'M'.         }
{----------------------------------------------------------------------------}
function MaxNorm( var M : rmatrix ) : real;
var
  i, j     : integer;
  Max, Tmp : real;
  AbsMi    : rvector[lb(M,2)..ub(M,2)];
begin
  Max := 0;
  for i := lb(M) to ub(M) do
  begin
    AbsMi := abs(M[i]);
    Tmp := #>( for j:=lb(M,2) to ub(M,2) sum( AbsMi[j] ) );
    if (Tmp > Max) then Max := Tmp;
  end;
  MaxNorm := Max;
end;

{----------------------------------------------------------------------------}
{ The vectors x and y are successive approximations for the solution of a    }
{ linear system of equations computed by iterative refinement. If a compo-   }
{ nent of y is diminished by more than 'Factor', it is a good candidate for  }
{ a zero entry. Thus, it is set to zero.                                     }
{----------------------------------------------------------------------------}
procedure CheckForZeros ( var x, y : rvector );
const
  Factor = 1E+5;
var
  i : integer;
begin
  for i := lb(y) to ub(y) do
    if ( abs(y[i])*Factor < abs(x[i]) ) then y[i] := 0;
end;

{----------------------------------------------------------------------------}
{ The vectors x and y are successive iterates. The function returns TRUE if  }
{ the relative error of all components x_i and y_i is <= 10^(-12), i.e. y_i  }
{ has about 12 correct decimals. If x_i or y_i vanishes, the relative error  }
{ is implicitly set to zero.                                                 }
{----------------------------------------------------------------------------}
function Accurate ( var x, y : rvector ) : boolean;
const
  Delta     = 1E-12;   { Relative error bound }
var
  ok     : boolean;
  i, n   : integer;
  abs_yi : real;
begin
  i := lb(y); n := ub(y);
  repeat
    if (sign(x[i])*sign(y[i]) <= 0) then      { Relative error set to 0 }
      ok := true
    else
      begin
        abs_yi := abs(y[i]);                  { Relative error > Delta? }
        ok := (abs(y[i] - x[i]) <= Delta * abs_yi );
      end;
    i := i+1;
  until (not ok) or (i > n);
  Accurate := ok;
end;

{----------------------------------------------------------------------------}
{ This procedure 'VerificationStep()' performs the iteration                 }
{ [y] = blow([x],Eps), [x] = [z] + [C]*[y] for k = 1,2,... until the new     }
{ iterate [x] lies in the interior of [y] or until the maximum number of     }
{ iterations is exceeded. The flag 'IsVerified' is set if an inclusion in    }
{ the interior could be established.                                         }
{----------------------------------------------------------------------------}
procedure VerificationStep ( var xx, zz     : ivector;
                             var C          : imatrix;
                             var IsVerified : boolean );
const
  MaxIter = 3;        { Maximum number of iteration steps }
  Epsilon = 1000;     { Factor for the epsilon inflation }
var
  p  : integer;
  yy : ivector[lb(xx)..ub(xx)];
begin
  xx := zz; p := 0;                       { Initialize:  [x] := [z] }
  repeat
    yy := blow(xx,Epsilon);               { Epsilon inflation               }
    xx := zz + C*yy;                      { New iterate: [x] := [z]+[C]*[y] }
    IsVerified := xx in yy;               { Inclusion in the interior?      }
    p := p+1;
  until ( IsVerified or (p >= MaxIter) );
end;

{----------------------------------------------------------------------------}
{ Purpose: The procedure 'LinSolveMain()' computes a verified solution of a  }
{    square linear system of equations A*x=b.                                }
{ Parameters:                                                                }
{    In  : 'A'           : matrix of the system, passed as reference         }
{                          parameter to save time for copying it.            }
{          'b'           : right-hand side of the system, passed as          }
{                          reference parameter to save time for copying it.  }
{          'ComputeCond' : flag signalling whether a condition number        }
{                          estimate is to be computed.                       }
{    Out : 'xx'          : enclosure of the unique solution.                 }
{          'Cond'        : condition number estimate.                        }
{          'Err'         : error code.                                       }
{ Description: An approximate inverse 'R' of 'A' is computed by calling      }
{   procedure 'MatInv()'. Then an approximate solution 'x' is computed       }
{   applying a conventional real residual iteration. For the final verifica- }
{   tion, an interval residual iteration is performed. An enclosure of the   }
{   unique solution is returned in the interval vector 'xx'. The procedure   }
{   also returns a condition number estimate 'Cond' if the flag 'ComputeCond'}
{   is set. 'Cond' is initialized by -1. A negative value for 'Cond' signals }
{   that an estimate could not be computed.                                  }
{----------------------------------------------------------------------------}
procedure LinSolveMain ( var A           : rmatrix; var b    : rvector;
                         var xx          : ivector; var Cond : real;
                             ComputeCond : boolean; var Err  : integer );
const
  MaxResCorr = 10;   { Maximum number of real residual corrections }

var
  IsVerified : boolean;
  k, n, m    : integer;
  x, y, d    : rvector[lb(xx)..ub(xx)];
  zz         : ivector[lb(xx)..ub(xx)];
  R          : rmatrix[lb(A,1)..ub(A,1),lb(A,2)..ub(A,2)];
  C          : imatrix[lb(A,1)..ub(A,1),lb(A,2)..ub(A,2)];

begin
  { Get length of the rows and columns of 'A', initialize condition number }
  n := ub(A,1) - lb(A,1);  m := ub(A,2) - lb(A,2);  Cond := -1;

  if (n <> m) then                        { Error: 'A' is not square     }
    Err := NotSquare
  else if ( (n <> ub(b)-lb(b)) or         { Error: Dimensions of A*x = b }
            (m <> ub(xx)-lb(xx)) ) then   {        are not compatible    }
    Err := DimensionErr
  else
    Err := NoError;

  if (Err = NoError) then
  begin
    MatInv(A,R,Err);
    if (Err <> NoError) then
      Err := InvFailed
    else
      begin {Algorithm}
        if ComputeCond then                       { Compute condition number}
          Cond := MaxNorm(A) *> MaxNorm(R);       {-------------------------}

        x := R*b;  k := 0;             { Real residual iteration }
        repeat                         {-------------------------}
          y := x;
          d := #*(b - A*y);
          x := #*(y + R*d);
          CheckForZeros(y,x);
          k := k+1;
        until Accurate(y,x) or (k >= MaxResCorr);

        { Prepare verification step, i.e. compute enclosures [C] }
        { and [z] of C = (I - R*A) and z = R*(b - A*x).          }
        {--------------------------------------------------------}
        C := ##(id(A) - R*A);
        d := #*(b - A*x);
        zz := ##(b - A*x - d);    { Now b-A*x lies in d+[z]. }
        zz := ##(R*d + R*zz);     { Thus, R*(b-A*x) lies in ##(R*d+R*[z]) }

        { If R*(b-A*x) = 0, then x is an exact solution. Otherwise try to }
        { find a verified enclosure [x] for the absolute error of x.      }
        {-----------------------------------------------------------------}
        if (zz = null(zz)) then
          xx := x
        else
          begin
            VerificationStep(xx,zz,C,IsVerified);   { Attempt to compute [x] }
            if not IsVerified then
              Err := VerifFailed
            else
              xx := x + xx;   { The exact solution lies in x+[x] }
          end;
      end; {Algorithm}
  end; { if (Err = NoError) ... }
end; {procedure LinSolve}

{----------------------------------------------------------------------------}
{ Purpose: The procedure 'LinSolve()' computes a verified solution of a      }
{    square linear system of equations A*x=b without returning a condition   }
{    number estimate.                                                        }
{ Parameters:                                                                }
{    In  : 'A'           : matrix of the system, passed as reference         }
{                          parameter to save time for copying it.            }
{          'b'           : right-hand side of the system, passed as          }
{                          reference parameter to save time for copying it.  }
{    Out : 'xx'          : enclosure of the unique solution.                 }
{          'Err'         : error code.                                       }
{ Description: Calls 'LinSolveMain()' for solving the linear system with the }
{    flag 'ComputeCond' not set.                                             }
{----------------------------------------------------------------------------}
global procedure LinSolve ( var A  : rmatrix; var b   : rvector;
                            var xx : ivector; var Err : integer );
var
  DummyCond : real;   { Dummy parameter for call of 'LinSolveMain()' }
begin
  LinSolveMain(A,b,xx,DummyCond,false,Err);
end;

{----------------------------------------------------------------------------}
{ Purpose: The procedure 'LinSolve()' computes a verified solution of a      }
{    square linear system of equations A*x=b and returns a condition         }
{    number estimate.                                                        }
{ Parameters:                                                                }
{    In  : 'A'           : matrix of the system, passed as reference         }
{                          parameter to save time for copying it.            }
{          'b'           : right-hand side of the system, passed as          }
{                          reference parameter to save time for copying it.  }
{    Out : 'xx'          : enclosure of the unique solution.                 }
{          'Cond'        : condition number estimate.                        }
{          'Err'         : error code.                                       }
{ Description: Calls 'LinSolveMain()' for solving the linear system with the }
{    flag 'ComputeCond' set.                                                 }
{----------------------------------------------------------------------------}
global procedure LinSolve ( var A   : rmatrix; var b    : rvector;
                            var xx  : ivector; var Cond : real;
                            var Err : integer                      );
begin
  LinSolveMain(A,b,xx,Cond,true,Err);
end;

{----------------------------------------------------------------------------}
{ Module initialization part                                                 }
{----------------------------------------------------------------------------}
begin
  { Nothing to initialize }
end.
